/*
 * AMPL interface for E04UCC (SQP for dense nonlinear programming problems)
 * intended for self-building by NAG C Library users
 * (Windows & Linux)
 *
 * version: 0.9 (beta)
 *
 * last update: 29/11/2010
 * by Jan Fiala, jan@nag.co.uk
 *
 *
 * Instructions:
 * (1) Prerequsities
 *     * ASL (AMPL Solver Library)
 *     * NAG C Library
 *
 *   If you don't have a suitable build of the ASL library you can
 *   download its source codes from
 *       http://netlib.sandia.gov/cgi-bin/netlib/netlibfiles.tar?filename=netlib/ampl/solvers
 *   as a tar archive. Or from
 *       http://www.netlib.org/ampl/solvers/
 *       http://netlib.sandia.gov/ampl/solvers/index.html
 *   as separate files. Please unzip it and follow README file for
 *   further instructions.
 *
 * (2) Building
 *   Choose appropriate Makefile according to your library and compiler.
 *   For example 
 *       Makefile.clw3209dal.cl
 *   is suitable if you use MS Visual C++ compiler (cl) and NAG C
 *   Library mark 9 for Windows (see comments inside). Edit library
 *   paths and other settings as desired. To build AMPL solvers invoke
 *   'make' on target 'all' as described inside the file depending on
 *   your system, for example:
 *       make -f Makefile.clw3209dal.gcc all
 *       nmake -f Makefile.clw3209dal.cl all
 *   If there is no Makefile for your specific combination of the system,
 *   library and compiler, edit a generic one as necessary.
 *
 * (3) Testing
 *   Invoke 'make' on target 'test' to run the embedded test, for example
 *       make -f Makefile.clw3209dal.gcc test
 *       nmake -f Makefile.clw3209dal.cl test
 *   By now your AMPL solver should work. In case of any troubles, please
 *   do not hesitate to contact NAG.
 *
 * Common problems:
 * (a) If you use dynamic link with the NAG Library don't forget to include
 * the library path to the system variables (PATH in Windows, LD_LIBRARY_PATH
 * in Linux). If there are any doubts if the library and/or licence is
 * accesible run the executable without any arguments. This will check the
 * licence and library calls. 
 * (b) When you call the solver from AMPL by "option solver solver_name;"
 * please do not forget that the solver is called directly by solver_name
 * thus 'solver_name' must include an absolute path or a relative path to
 * the current location. For example
 *     option solver e04ug;
 * (if you are on Windows in the same directory as the solver)
 *     option solver "./e04ug.exe";
 * (if you are on Linux in the same directory as the solver but the directory
 * is not included in your PATH variable)
 *
 *
 * NAG Lib calls: 
 *    nag_licence_query (a00acc)
 *    [nag_implementation_details (a00aac)]
 *    nag_opt_init (e04xxc)
 *    nag_opt_nlp (e04ucc)
 *    nag_opt_free (e04xzc)
 */

#include <stdio.h>
#include <malloc.h>

// AMPL interface library (their IO)
// [needs to go first because of NAG redefinitions of Vsprintf etc.]
#include "stdio1.h"

// declarations of NAG C Library subroutines, data types, error messages etc.
#include <nag.h>
#include <nag_stdlib.h>
#include <naga00.h>
#include <nage04.h>

// AMPL interface library (the rest)
#include "asl.h"
#include "getstub.h"

// verify derivatives? No, not necessary, are exact (otherwise comment it out)
#define DONOTVERIFY

// definition of keywords to modify default option settings of E04UCC
typedef struct {
  double tol_feas;      // Feasibility Tolerance
  double tol_prec;      // Function Precision
  double tol_optim;     // optimaliry tolerance
  int maxit;            // Major Iteration Limit
  int printlev;         // Print Level
} e04ucc_parameters;

// callback functions declarations
#ifdef __cplusplus
extern "C" {
#endif
static void NAG_CALL objfun(Integer n, double x[], double *objf,
                            double objgrd[], Nag_Comm *comm);
static void NAG_CALL confun(Integer n, Integer ncnlin, Integer needc[],
                            double x[], double conf[], double conjac[],
                            Nag_Comm *comm);
#ifdef __cplusplus
}
#endif

// Ampl Solver Library (ASL) data structure
// (doesn't need to be global but it makes life easier...)
static ASL *asl=0;

// to store any deviations from default opt. settings, -1 to indicate no change
static e04ucc_parameters param = {-1, -1, -1, -1, -1}; 

// definition of keywords AMPL should react to
static keyword keywds[] = {
    KW("feasibility",D_val, &param.tol_feas,	"Feasibility Tolerance (default sqrt(eps))"),
    KW("maxit",	     I_val, &param.maxit, "Major Iteration Limit (default based on the problem size)"),
    KW("optimality", D_val, &param.tol_optim,	"Optimality Tolerance (default eps^0.8)"),
    KW("precision",	 D_val, &param.tol_prec,	"Function Precision (default eps^0.9)"),
    KW("print",      I_val, &param.printlev, "Print Level (0..7, default 4)"),
  };

// usage message to be printed for '-?'
static char *usage_msg[] = {
  "  where  stub  is a file created from  `ampl -obstub`  or  `ampl -ogstub`",
  "  call.  Assignments have the form  `keyword=value`. To get a list of all",
  "  keywords, call `./e04uc.exe -=`, see description bellow.",
  "  Example:",
  "    ./e04uc.exe problem.nl maxit=100 precision=1e-2 print=0",
  " ",
  "  Alternatively, you can invoke solver E04UCC from Ampl environment. Run",
  "  `ampl` and input the following commands:",
  "     model problem.mod;",
  "     option solver e04uc;",
  "     option e04uc_options 'maxit=100 precision=1e-2 print=1';",
  "     solve;",
  0 };

// possible final messages
static char *end_msg[] = {
  /*0*/ "converged: optimal solution found",
  /*1*/ "converged: suboptimal solution found, accuracy couldn't be achieved",
  /*2*/ "failed: the problem is infeasible (linear constraints & bounds)",
  /*3*/ "failed: the problem is infeasible (nonlinear constraints)",
  /*4*/ "failed: maximum number of iterations reached",
  /*5*/ "failed: requested accuracy couldn't be achieved",
  /*6*/ "error: error in the option settings",
  /*7*/ "error: please consult the manual"
  };

// info structure of the solver, messages, option settings
static Option_Info Oinfo = {"E04UC", "E04UC, NAG C Library","e04uc_options", keywds, nkeywds, 0 /*funcadd?*/, "E04UC, NAG C Library, AMPL interface 0.9 (beta)", usage_msg, 0, 0, 0 /*cmd line options?*/, 0, 20101118};

// set non-default option settings given by user through AMPL
void set_options(Nag_E04_Opt *options) {
  const int prnlevels[]={Nag_NoPrint, Nag_Soln, Nag_Iter, Nag_Iter_Long, Nag_Soln_Iter, Nag_Soln_Iter_Long, Nag_Soln_Iter_Const, Nag_Soln_Iter_Full};

  if (!options)
    return;

#ifdef DONOTVERIFY
  // turn off the verification of derivatives, they are computed via
  // automatic differentiation so they must be exact
  options->verify_grad=Nag_NoCheck;
#endif

  if (param.tol_feas>0) {
    options->nonlin_feas_tol=param.tol_feas;
    options->lin_feas_tol=param.tol_feas;
  }
  if (param.tol_prec>0)
    options->f_prec=param.tol_prec;
  if (param.tol_optim>0)
    options->optim_tol=param.tol_optim;
  if (param.maxit>=0)
    options->max_iter=param.maxit;
  if (param.printlev>=0 && param.printlev<=7)
    options->print_level=prnlevels[param.printlev];
  else if (param.printlev>7)
    options->print_level=prnlevels[7];

}

// Evaluate the objective function & 1st derivatives
// expects: comm->iuser[0]=nobj, comm->iuser[1]=objsign
static void NAG_CALL objfun(Integer n, double x[], double *objf, double objgrd[], Nag_Comm *comm) {

  int nobj, objsign, i;
  fint nerror=0;  // change it to longjumps error treatment if needed

  if (!comm)
    return;
  if (!x || !objf || !objgrd || !comm->iuser) {
    comm->flag=-1;
    return;
  }
  nobj=comm->iuser[0];
  objsign=comm->iuser[1];

  // return objective value
  if (comm->flag == 0 || comm->flag == 2) {
    if (nobj>=0) {
      *objf=objsign * objval(nobj,x,&nerror);
      if (nerror) {
        printf("Error while evaluating objective function, terminating.\n");
        comm->flag=-1;
        return;
      }
    }
    else  // const. objective
      *objf=0;
  }

  // return objective derivative
  if (comm->flag == 2) {
    if (nobj>=0) {
      objgrd(nobj, x, objgrd, &nerror);
      for (i=0;i<n;i++)
        objgrd[i]*=objsign;
      if (nerror) {
        printf("Error while evaluating objective function, terminating.\n");
        comm->flag=-1;
      }
    }
    // else nothing necessary, was zero vector, will be a zero vector
  }

}

// evaluate constraints and their derivative
static void NAG_CALL confun(Integer n, Integer ncnlin, Integer needc[], double x[], double conf[], double conjac[], Nag_Comm *comm) {
  int i;
  fint nerror=0;  // change it to longjumps error treatment if needed

  if (!comm)
    return;
  if (!needc || !x || !conf || !conjac) {
    comm->flag=-1;
    return;
  }

  // return constraints values
  if (comm->flag == 0 || comm->flag == 2)
    for (i=0; i<ncnlin; i++)
      if (needc[i]>0) {
        conf[i]=conival(i,x,&nerror);  // nonlin constr. go first in AMPL so no need to change 'i'
        if (nerror) {
          printf("Error while evaluating constraints, terminating.\n");
          comm->flag=-1;
          return;
        }
      }
  

  // return constraints derivatives
  if (comm->flag == 2)
    for (i=0; i<ncnlin; i++)
      if (needc[i]>0) {
        // need to fill in conjac[i*n+j], j=0,..,n-1
        congrd(i,x,conjac+i*n,&nerror);
        if (nerror) {
          printf("Error while evaluating constraints, terminating.\n");
          comm->flag=-1;
          return;
        }
      }

}

int main(int argc, char *argv[]) {

  FILE *nlfile=0;
  char *stubname;
  char *progname=argv[0];
  int nobj=0;   // number of objective to use, -1 for constant
  int errflag;
  int retflag=0;
  int i;
  cgrad *cg;
  int objsign=1;
  int ifail, offset;

  // e04ucc vars
  double *a=0, *bl=0, *bu=0, *x=0, *g=0, objf, *ruser=0;
  Integer n, nclin, ncnlin, *iuser=0, tda;
  Nag_E04_Opt options;
  Nag_Comm comm;
  NagError fail;

  // prepare our AMPL problem structure to compute function values and 1st deriv
  asl = ASL_alloc(ASL_read_fg);
  // find problem name (stub) & react on -=, -?
  stubname = getstub(&argv, &Oinfo);
  if (!stubname) {
    printf("%s\n",Oinfo.bsname);
    printf("Licence: %s\n", nag_licence_query() ? "valid" : "valid licence NOT found");
    /*  With NAG_DLL it creates a pop-up window instead of the text to stdout
     *  so let's not use it
    printf("\nBased on library:\n");
    nag_implementation_details();
    */
    printf("\n");
    printf("No test problem set, terminating.\n");
    printf("Call `%s -?` for help.\n",progname);
    retflag=0;
    goto cleaning;
  }

  return_nofile = 1;
  // load basic characteristics of the problem
  nlfile = jac0dim(stubname, (ftnlen)strlen(stubname));
  if (!nlfile) {
    printf("Can't open %s\nTerminating.\n",stubname);
    retflag=1;
    goto cleaning;
  }

  // is it a "normal" nonlinear continuous problem?
  if (n_var<=0 || nbv>0 || niv>0 || nlnc>0 || lnc>0) {
    printf("Discrete problems or problems with newtork constraints are not supported by this solver.\nTerminating.\n");
    retflag=1;
    goto cleaning;
  }
  if (n_obj<=0) {
    printf("Warning: There is no objective function, pretending to be constant.\n");
    nobj=-1;
  } else if (n_obj>1) {
    printf("Warning: More than one objetive function detected, using the first one.\n");
  }
  
  // AMPL uses their "own" Malloc and type real - only to allocate their bits
  // by allocating these we notify AMPL to fill the data in
  // there are many macros in asl.h which changes many elements of the
  // structure 'asl' such that they look as local variables, for example:
  //    #define n_var   asl->i.n_var_
  //    #define LUv     asl->i.LUv_
  // So if there is no local definition of the variable it is likely to be
  // a part of 'asl' structure
  LUv = (real *)Malloc(n_var*sizeof(real));
  Uvx = (real *)Malloc(n_var*sizeof(real));
  LUrhs = (real *)Malloc(n_con*sizeof(real));
  Urhsx = (real *)Malloc(n_con*sizeof(real));
  X0 = (real *)Malloc(n_var*sizeof(real));
  pi0 = (real *)Malloc(n_con*sizeof(real));
  if (!X0 || !LUv || ! Uvx || (n_con && (!LUrhs || !Urhsx || !pi0))) {
    printf("Memory allocation problem, the problem is too big to fit into the memory. Terminating.\n");
    retflag=1;
    goto cleaning;
  }
  
  // read in the rest of the problem
  if ((errflag=fg_read(nlfile, ASL_return_read_err))!=0) {
    printf("AMPL problem with reading the problem, err_code: %i\nTerminating\n",errflag);
    retflag=1;
    goto cleaning;
  }

  // read additional options and store them into global 'param'
  getopts(argv, &Oinfo);  
  // initialize nag options structure and set it accordingly
  nag_opt_init(&options);
  set_options(&options);

  // correct min/max: AMPL: objtype[]==0 -> min, ==1 ->max
  if (nobj!=-1 && objtype[nobj])  
    objsign=-1;

  // E04UCC init
  n=n_var;
  nclin=n_con-nlc;  //# linear constr.
  ncnlin=nlc;       //# nonlin. constr.
  tda=n;

  a = (double *) calloc((nclin*tda + 1), sizeof(double)); // +1 in case nclin==0
  bl = (double *) malloc((n+nclin+ncnlin)*sizeof(double));
  bu = (double *) malloc((n+nclin+ncnlin)*sizeof(double));
  x = (double *) malloc(n*sizeof(double));
  g = (double *) calloc(n,sizeof(double));
  iuser = (Integer *) calloc(3,sizeof(Integer));
  ruser = 0; //(double *) malloc(2*n*sizeof(double));
  if (!a || !bl || !bu || !x || !g || !iuser) {
    printf("Memory allocation problem. The problem cannot fit in the memory as dense.\nTry a sparse routine instead.\nTerminating.\n");
    retflag=1;
    goto cleaning;
  }

  for (i=0; i<n; i++)
    x[i]=X0[i];
  // let's hope that infty for AMPL works well for bigbnd
  for (i=0; i<n;i++) {
    bl[i]=LUv[i];
    bu[i]=Uvx[i];
  }
  // first linear then nonlinear constraints not as in AMPL
  for (i=n,offset=nlc-n; i<n+nclin; i++) {  
    bl[i]=LUrhs[offset+i];
    bu[i]=Urhsx[offset+i];
  }
  for (i=n+nclin,offset=-n-nclin; i<n+nclin+ncnlin; i++) {
    bl[i]=LUrhs[offset+i];
    bu[i]=Urhsx[offset+i];
  }
  // filling in A - just nonzeros, anywhere else are zeros
  for (i=0; i<nclin; i++) {
    for (cg=Cgrad[nlc+i]; cg; cg=cg->next) {
      a[cg->varno + i*tda]=cg->coef;  // varno is 0-based -> OK
    }
  }

  iuser[0]=nobj;
  iuser[1]=objsign;
  comm.iuser=iuser;
  comm.user=ruser;
  objf=0.0;
  INIT_FAIL(fail);

  // optimize
  nag_opt_nlp(n, nclin, ncnlin, a, tda, bl, bu, objfun, confun, x, &objf,
              g, &options, &comm, &fail);

  // results check
  switch(fail.code) {
    case NE_NOERROR:
      // OK, solution found
      ifail=0;
      printf("Result OK\n");
      break;

    case NW_NOT_CONVERGED:
      // suboptimal solution found, accuracy not achieved
      ifail=1;
      printf("Warning from nag_opt_nlp (e04ucc).\n%s\n", fail.message);
      break;

    case NW_LIN_NOT_FEASIBLE:
      // infeasible problems
      ifail=2;
      printf("Warning from nag_opt_nlp (e04ucc).\n%s\n", fail.message);
      break;

    case NW_NONLIN_NOT_FEASIBLE:
      // infeasible problems
      ifail=3;
      printf("Warning from nag_opt_nlp (e04ucc).\n%s\n", fail.message);
      break;

    case NW_TOO_MANY_ITER:
      // maxit
      ifail=4;
      printf("Warning from nag_opt_nlp (e04ucc).\n%s\n", fail.message);
      break;

    case NW_KT_CONDITIONS:
      // KKT not satisfied, cannot be improved
      ifail=5;
      printf("Warning from nag_opt_nlp (e04ucc).\n%s\n", fail.message);
      break;

    case NE_2_INT_OPT_ARG_CONS:
    case NE_BAD_PARAM:
    case NE_INT_OPT_ARG_GT:
    case NE_INT_OPT_ARG_LT:
    case NE_INVALID_INT_RANGE_1:
    case NE_INVALID_REAL_RANGE_EF:
    case NE_INVALID_REAL_RANGE_F:
    case NE_INVALID_REAL_RANGE_FF:
      // wrong option settings
      ifail=6;
      printf("Error in option settings of nag_opt_nlp (e04ucc).\n%s\n", fail.message);
      break;

    // all errors reported to AMPL as 'general error, consult the manual'
    case NW_OVERFLOW_WARN:
      // serious ill-conditioning
    case NE_USER_STOP:
    case NE_ALLOC_FAIL:
      ifail=7;
      printf("Error from nag_opt_nlp (e04ucc).\n%s\n", fail.message);
      break;

    // these errors shouldn't happen - my coding errors or something strange
    case NE_2_INT_ARG_LT:
    case NE_BOUND:
    case NE_BOUND_EQ:
    case NE_BOUND_EQ_LCON:
    case NE_BOUND_EQ_NLCON:
    case NE_BOUND_LCON:
    case NE_BOUND_NLCON:
    case NE_DERIV_ERRORS:
    case NE_INT_ARG_LT:
    case NE_OPT_NOT_INIT:
    case NE_STATE_VAL:
      // my coding problem from AMPL
    case NE_NOT_APPEND_FILE:
    case NE_NOT_CLOSE_FILE:
    case NE_WRITE_ERROR:
    default:
      // strange errors which shouldn't happen
      ifail=7;
      printf("ERROR from nag_opt_nlp (e04ucc).\n%s\n", fail.message);
      break;
  }

  // write results to AMPL
  for (i=0;i<n_var;i++)
    X0[i]=x[i];
  // need to shuffle multipliers back (if they exist)
  if (options.lambda) {
    for (i=0;i<ncnlin;i++)
      pi0[i]=options.lambda[n+nclin+i];
    for (;i<n_con;i++)
      pi0[i]=options.lambda[n+i-ncnlin];
  } else
    for (i=0;i<n_con;i++)
      pi0[i]=0;

  // report the solution back to AMPL
  write_sol( (ifail>=0 && ifail<7) ? end_msg[ifail] : end_msg[7],X0,pi0,&Oinfo);

  // free all the option settings
  INIT_FAIL(fail);
  nag_opt_free(&options, "all", &fail);

cleaning: ;  // something to clean?
  if (a) free(a);
  if (bl) free(bl);
  if (bu) free(bu);
  if (x) free(x);
  if (g) free(g);
  if (iuser) free(iuser);
  if (ruser) free(ruser);

  if (asl)
    ASL_free(&asl);
      
  return retflag;
}

