/*
 * AMPL interface for E04UFF (SQP for dense nonlinear programming problems)
 * intended for self-building by NAG Fortran Library users
 * (Windows & Linux)
 *
 * version: 0.9 (beta)
 *
 * last update: 29/11/2010
 * by Jan Fiala, jan@nag.co.uk
 *
 *
 * Instructions:
 * (1) Prerequsities
 *     * ASL (AMPL Solver Library)
 *     * NAG Fortran Library
 *
 *   If you don't have a suitable build of the ASL library you can
 *   download its source codes from
 *       http://netlib.sandia.gov/cgi-bin/netlib/netlibfiles.tar?filename=netlib/ampl/solvers
 *   as a tar archive. Or from
 *       http://www.netlib.org/ampl/solvers/
 *       http://netlib.sandia.gov/ampl/solvers/index.html
 *   as separate files. Please unzip it and follow README file for
 *   further instructions.
 *
 * (2) Building
 *   Choose appropriate Makefile according to your library and compiler.
 *   For example 
 *       Makefile.fldll224ml.cl
 *   is suitable if you use MS Visual C++ compiler (cl) and NAG Fortran
 *   Library mark 22 Windows DLL (see comments inside). Edit library
 *   paths and other settings as desired. To build AMPL solvers invoke
 *   'make' on target 'all' as described inside the file depending on
 *   your system, for example:
 *       make -f Makefile.fldll224ml.gcc all
 *       nmake -f Makefile.fldll224ml.cl all
 *   If there is no Makefile for your specific combination of the system,
 *   library and compiler, edit a generic one as necessary.
 *
 * (3) Testing
 *   Invoke 'make' on target 'test' to run the embedded test, for example
 *       make -f Makefile.fldll224ml.gcc test
 *       nmake -f Makefile.fldll224ml.cl test
 *   By now your AMPL solver should work. In case of any troubles, please
 *   do not hesitate to contact NAG.
 *
 * Common problems:
 * (a) If you use dynamic link with the NAG Library don't forget to include
 * the library path to the system variables (PATH in Windows, LD_LIBRARY_PATH
 * in Linux). If there are any doubts if the library and/or licence is
 * accesible run the executable without any arguments. This will check the
 * licence and library calls. 
 * (b) When you call the solver from AMPL by "option solver solver_name;"
 * please do not forget that the solver is called directly by solver_name
 * thus 'solver_name' must include an absolute path or a relative path to
 * the current location. For example
 *     option solver e04ug;
 * (if you are on Windows in the same directory as the solver)
 *     option solver "./e04ug.exe";
 * (if you are on Linux in the same directory as the solver but the directory
 * is not included in your PATH variable)
 *
 *
 * NAG Lib calls: E04UEF, E04UFF, A00AAF, A00ACF
 */

#include <stdio.h>
#include <malloc.h>

// declarations of NAG Fortran Library subroutines
#include "flnage04.h"

// AMPL interface library
#include "asl.h"
#include "getstub.h"

// verify derivatives? No, not necessary, are exact
#define VERIFYLEVEL     -1

// definition of keywords to modify default option settings of E04UFF
typedef struct {
  double tol_feas;      // Feasibility Tolerance
  double tol_prec;      // Function Precision
  double tol_optim;     // optimaliry tolerance
  int maxit;            // Major Iteration Limit
  int printlev;         // Print Level
} e04uff_parameters;

// Ampl Solver Library (ASL) data structure
// (doesn't need to be global but it makes life easier...)
static ASL *asl=0;  

// to store any deviations from default opt. settings, -1 to indicate no change
static e04uff_parameters param = {-1, -1, -1, -1, -1}; 

// definition of keywords AMPL should react to
static keyword keywds[] = {
    KW("feasibility",D_val, &param.tol_feas,	"Feasibility Tolerance (default sqrt(eps))"),
    KW("maxit",	     I_val, &param.maxit, "Major Iteration Limit (default auto)"),
    KW("optimality", D_val, &param.tol_optim,	"Optimality Tolerance (default eps^0.8)"),
    KW("precision",	 D_val, &param.tol_prec,	"Function Precision (default eps^0.9)"),
    KW("print",      I_val, &param.printlev, "Print Level (default 10)"),
  };

// usage message to be printed for '-?'
static char *usage_msg[] = {
  "  where  stub  is a file created from  `ampl -obstub`  or  `ampl -ogstub`",
  "  call.  Assignments have the form  `keyword=value`. To get a list of all",
  "  keywords, call `./e04uf.exe -=`, see description bellow.",
  "  Example:",
  "    ./e04uf.exe problem.nl maxit=100 precision=1e-2 print=0",
  " ",
  "  Alternatively, you can invoke solver E04UFF from Ampl environment. Run",
  "  `ampl` and input the following commands:",
  "     model problem.mod;",
  "     option solver e04uf;",
  "     option e04uf_options 'maxit=100 precision=1e-2 print=1';",
  "     solve;",
  0 };

// possible final messages
static char *end_msg[] = {
  /*0*/ "converged: optimal solution found",
  /*1*/ "converged: suboptimal solution found, accuracy couldn't be achieved",
  /*2*/ "failed: the problem is infeasible (linear constraints & bounds)",
  /*3*/ "failed: the problem is infeasible (nonlinear constraints)",
  /*4*/ "failed: maximum number of iterations reached",
  /*5*/ "error: please consult the manual",
  /*6*/ "failed: requested accuracy couldn't be achieved",
  /*7*/ "error: please consult the manual"
  };

// info structure of the solver, messages, option settings
static Option_Info Oinfo = {"E04UF", "E04UF, NAG Fortran Library","e04uf_options", keywds, nkeywds, 0 /*funcadd?*/, "E04UF, NAG Fortran Library, AMPL interface 0.9 (beta)", usage_msg, 0, 0, 0 /*cmd line options?*/, 0, 20101111};

// set non-default option settings given by user through AMPL
void set_options(void) {
  char buff[80];
  int len=80;

  // set the level for verification of derivatives, to turn off -> -1
  len=sprintf(buff,"Verify Level = %i ",VERIFYLEVEL);
  E04UEF(buff,len);

  if (param.tol_feas>0) {
    len=sprintf(buff,"Feasibility Tolerance = %e ",param.tol_feas);
    E04UEF(buff,len);
  }
  if (param.tol_prec>0) {
    len=sprintf(buff,"Function Precision = %e ",param.tol_prec);
    E04UEF(buff,len);
  }
  if (param.tol_optim>0) {
    len=sprintf(buff,"Optimality Tolerance = %e ",param.tol_optim);
    E04UEF(buff,len);
  }
  if (param.maxit!=-1) {
    len=sprintf(buff,"Iteration Limit = %i ",param.maxit);
    E04UEF(buff,len);
  }
  if (param.printlev!=-1) {
    len=sprintf(buff,"Print Level = %i ",param.printlev);
    E04UEF(buff,len);
  }
}

int main(int argc, char *argv[]) {
  FILE *nlfile=0;
  char *stubname;
  char *progname=argv[0];
  int nobj=0;   // number of objective to use, -1 for constant
  int errflag;
  int retflag=0;
  int i,j,offset;
  real *vec=0;
  fint nerror=0;  // change it to longjumps error treatment if needed
  cgrad *cg;
  int objsign=1;
  
  // e04uff vars
  int irevcm, n, nclin, ncnln, lda, ldcj, ldr, iter, *istate=0, *needc=0, *iwork=0, liwork, lwork, ifail;
  double *a=0, *bl=0, *bu=0, *c=0, *cjac=0, *clambda=0, objf, *objgrd=0, *r=0, *x=0, *work=0;


  // prepare our AMPL problem structure to compure function values and 1st deriv
  asl = ASL_alloc(ASL_read_fg);
  // find problem name (stub) & react on -=, -?
  stubname = getstub(&argv, &Oinfo);  
  if (!stubname) {
    printf("%s\n",Oinfo.bsname);
    printf("Licence: %s\n\nBased on library:\n", A00ACF() ? "valid" : "valid licence NOT found");
    A00AAF();   // print info about NAG library
    printf("\n");
    printf("No test problem set, terminating.\n");
    printf("Call `%s -?` for help.\n",progname);
    retflag=0;
    goto cleaning;
  }

  return_nofile = 1;
  // load basic characteristics of the problem
  nlfile = jac0dim(stubname, (ftnlen)strlen(stubname));
  if (!nlfile) {
    printf("Can't open %s\nTerminating.\n",stubname);
    retflag=1;
    goto cleaning;
  }

  // is it a "normal" nonlinear continuous problem?
  if (n_var<=0 || nbv>0 || niv>0 || nlnc>0 || lnc>0) {
    printf("Discrete problems or problems with newtork constraints are not supported by this solver.\nTerminating.\n");
    retflag=1;
    goto cleaning;
  }
  if (n_obj<=0) {
    printf("Warning: There is no objective function, pretending to be constant.\n");
    nobj=-1;
  } else if (n_obj>1) {
    printf("Warning: More than one objetive function detected, using the first one.\n");
  }

  // AMPL uses their "own" Malloc and type real - only to allocate their bits
  // by allocating these we notify AMPL to fill the data in
  // there are many macros in asl.h which changes many elements of the
  // structure 'asl' such that they look as local variables, for example:
  //    #define n_var   asl->i.n_var_
  //    #define LUv     asl->i.LUv_
  // So if there is no local definition of the variable it is likely to be
  // a part of 'asl' structure
  LUv = (real *)Malloc(n_var*sizeof(real));
  Uvx = (real *)Malloc(n_var*sizeof(real));
  LUrhs = (real *)Malloc(n_con*sizeof(real));
  Urhsx = (real *)Malloc(n_con*sizeof(real));
  X0 = (real *)Malloc(n_var*sizeof(real));
  pi0 = (real *)Malloc(n_con*sizeof(real));
  if (!X0 || !LUv || ! Uvx || (n_con && (!LUrhs || !Urhsx || !pi0))) {
    printf("Memory allocation problem, the problem is too big to fit into the memory. Terminating.\n");
    retflag=1;
    goto cleaning;
  }
  
  // read the rest of the problem in
  if ((errflag=fg_read(nlfile, ASL_return_read_err))!=0) {
    printf("AMPL problem with reading the problem, err_code: %i\nTerminating\n",errflag);
    retflag=1;
    goto cleaning;
  }

  // read additional options and store them into global 'param'
  getopts(argv, &Oinfo);
  set_options();

  // correct min/max: AMPL: objtype[]==0 -> min, ==1 ->max
  if (nobj!=-1 && objtype[nobj])  
    objsign=-1;

  // E04UFF init
  irevcm=0;
  n=n_var;
  nclin=n_con-nlc;  //# linear constr.
  ncnln=nlc;        //# nonlin. constr.
  lda=nclin ? nclin : 1;
  ldcj=ncnln ? ncnln : 1;
  ldr=n;
  iter=0;
  liwork=3*n+nclin+2*ncnln;
  lwork=2*n*n+n*nclin+2*n*ncnln+21*n+11*nclin+22*ncnln+1;
  ifail=-1;
  a = (double *) calloc((nclin*n + 1), sizeof(double));  // +1 in case nclin==0
  bl = (double *) malloc((n+nclin+ncnln)*sizeof(double));
  bu = (double *) malloc((n+nclin+ncnln)*sizeof(double));
  istate = (int *) calloc((n+nclin+ncnln), sizeof(int));
  c = (double *) calloc((ncnln + 1), sizeof(double));
  cjac = (double *) calloc((ncnln*n + 1), sizeof(double));
  clambda = (double *) calloc((n+nclin+ncnln), sizeof(double));
  objgrd = (double *) calloc(n, sizeof(double));
  r = (double *) calloc(n*n, sizeof(double));
  x = (double *) malloc(n*sizeof(double));
  needc = (int *) calloc((ncnln+1), sizeof(int));
  iwork = (int *) calloc(liwork, sizeof(int));
  work = (double *) calloc(lwork, sizeof(double));
  vec =  (real *) malloc(n_var*sizeof(real));
  
  if (!a || !bl || !bu || !istate || !c || !cjac || !clambda || !objgrd || !r || !x || !needc || !iwork || !work || !vec) {
    printf("Memory allocation problem. The problem cannot fit in the memory as dense.\nTry a sparse routine instead.\nTerminating.\n");
    retflag=1;
    goto cleaning;
  }

  for (i=0; i<n; i++)
    x[i]=X0[i];
  // let's hope that infty for AMPL works well for bigbnd
  for (i=0; i<n;i++) {
    bl[i]=LUv[i];
    bu[i]=Uvx[i];
  }
  for (i=n,offset=nlc-n; i<n+nclin; i++) {  
    // don't forget correction of the absolute term for linear!
    bl[i]=LUrhs[offset+i];
    bu[i]=Urhsx[offset+i];
  }
  for (i=n+nclin,offset=-n-nclin; i<n+nclin+ncnln; i++) {
    bl[i]=LUrhs[offset+i];
    bu[i]=Urhsx[offset+i];
  }
  // filling in A - just nonzeros, anywhere else are zeros
  for (i=0; i<nclin; i++) {
    for (cg=Cgrad[nlc+i]; cg; cg=cg->next) {
      a[cg->varno*nclin+i]=cg->coef;  // varno is 0-based, column-wise storage
    }
  }

  // reverse communication cycle
  do {
    E04UFF(&irevcm, &n, &nclin, &ncnln, &lda, &ldcj, &ldr, a, bl, bu,
           &iter, istate, c, cjac, clambda, &objf, objgrd, r, x,
           needc, iwork, &liwork, work, &lwork, &ifail);
    switch (irevcm) {
    case 1:
    case 2:
    case 3:
      if (irevcm & 1) {  // eval objf
        // don't forget constant objective!
        if (nobj>=0) {
          objf=objsign * objval(nobj,x,&nerror);
          if (nerror) {
            printf("Error while evaluating objective function, terminating.\n");
            irevcm=-1;
          }
        }
        else  // const. objective
          objf=0;
      }
      if (irevcm & 2) {  // eval objgrd
        if (nobj>=0) {
          objgrd(nobj, x, objgrd, &nerror);
          for (i=0;i<n;i++)
            objgrd[i]*=objsign;
          if (nerror) {
            printf("Error while evaluating objective function, terminating.\n");
            irevcm=-1;
          }
        }
        // else nothing necessary, was zero vector, will be a zero vector
      }
      break;
    case 4:
    case 5:
    case 6:
      if (irevcm==4 || irevcm==6) { // eval c(i) if needc[i]>0
        for (i=0; i<ncnln; i++)
          if (needc[i]>0) {
            c[i]=conival(i,x,&nerror);  // nonlin constr. go first in AMPL so no need to change 'i'
            if (nerror) {
              printf("Error while evaluating constraints, terminating.\n");
              irevcm=-1;
              break;
            }
          }
      }
      if (irevcm==5 || irevcm==6) { // eval c'(i) same
        for (i=0; i<ncnln; i++)
          if (needc[i]>0) {
            congrd(i,x,vec,&nerror);
            for (j=0;j<n;j++) {
              cjac[j*ncnln+i]=vec[j];
            }
            if (nerror) {
              printf("Error while evaluating constraints, terminating.\n");
              irevcm=-1;
              break;
            }
          }
      }
      break;
    }
  } while (irevcm);

  // results check & save/store results
  if (ifail)
    printf("E04UF finished with an error/warning; flag: %i\n",ifail);
  else
    printf("Result OK\n");

  // write results to AMPL
  for (i=0;i<n_var;i++)
    X0[i]=x[i];
  // need to shuffle multipliers back
  for (i=0;i<ncnln;i++)
    pi0[i]=clambda[n+nclin+i];
  for (;i<n_con;i++)
    pi0[i]=clambda[n+i-ncnln];
  // report the solution back to AMPL
  write_sol( (ifail>=0 && ifail<7) ? end_msg[ifail] : end_msg[7],X0,pi0,&Oinfo);

cleaning: ;  // something to clean?
  if (a) free(a);
  if (bl) free(bl);
  if (bu) free(bu);
  if (c) free(c);
  if (cjac) free(cjac);
  if (clambda) free(clambda);
  if (objgrd) free(objgrd);
  if (r) free(r);
  if (x) free(x);
  if (work) free(work);
  if (istate) free(istate);
  if (needc) free(needc);
  if (iwork) free(iwork);
  if (vec) free(vec);

  if (asl)
    ASL_free(&asl);
      
  return retflag;
}

