/*
 * AMPL interface for E04UGF (SQP for sparse nonlinear programming problems)
 * intended for self-building by NAG Fortran Library users
 * (Windows & Linux)
 *
 * version: 0.9 (beta)
 *
 * last update: 29/11/2010
 * by Jan Fiala, jan@nag.co.uk
 *
 *
 * Instructions:
 * (1) Prerequsities
 *     * ASL (AMPL Solver Library)
 *     * NAG Fortran Library
 *
 *   If you don't have a suitable build of the ASL library you can
 *   download its source codes from
 *       http://netlib.sandia.gov/cgi-bin/netlib/netlibfiles.tar?filename=netlib/ampl/solvers
 *   as a tar archive. Or from
 *       http://www.netlib.org/ampl/solvers/
 *       http://netlib.sandia.gov/ampl/solvers/index.html
 *   as separate files. Please unzip it and follow README file for
 *   further instructions.
 *
 * (2) Building
 *   Choose appropriate Makefile according to your library and compiler.
 *   For example 
 *       Makefile.fldll224ml.cl
 *   is suitable if you use MS Visual C++ compiler (cl) and NAG Fortran
 *   Library mark 22 Windows DLL (see comments inside). Edit library
 *   paths and other settings as desired. To build AMPL solvers invoke
 *   'make' on target 'all' as described inside the file depending on
 *   your system, for example:
 *       make -f Makefile.fldll224ml.gcc all
 *       nmake -f Makefile.fldll224ml.cl all
 *   If there is no Makefile for your specific combination of the system,
 *   library and compiler, edit a generic one as necessary.
 *
 * (3) Testing
 *   Invoke 'make' on target 'test' to run the embedded test, for example
 *       make -f Makefile.fldll224ml.gcc test
 *       nmake -f Makefile.fldll224ml.cl test
 *   By now your AMPL solver should work. In case of any troubles, please
 *   do not hesitate to contact NAG.
 *
 * Common problems:
 * (a) If you use dynamic link with the NAG Library don't forget to include
 * the library path to the system variables (PATH in Windows, LD_LIBRARY_PATH
 * in Linux). If there are any doubts if the library and/or licence is
 * accesible run the executable without any arguments. This will check the
 * licence and library calls. 
 * (b) When you call the solver from AMPL by "option solver solver_name;"
 * please do not forget that the solver is called directly by solver_name
 * thus 'solver_name' must include an absolute path or a relative path to
 * the current location. For example
 *     option solver e04ug;
 * (if you are on Windows in the same directory as the solver)
 *     option solver "./e04ug.exe";
 * (if you are on Linux in the same directory as the solver but the directory
 * is not included in your PATH variable)
 *
 *
 * NAG Lib calls: E04UJF, E04UGF, A00AAF, A00ACF
 */

#include <stdio.h>
#include <malloc.h>

// declarations of NAG Fortran Library subroutines
#include "flnage04.h"

// AMPL interface library
#include "asl.h"
#include "getstub.h"

// verify derivatives? No, not necessary, are exact
#define VERIFYLEVEL     -1

// definition of keywords to modify default option settings of E04UGF
typedef struct {
  double tol_feas;   // Feasibility Tolerance
  double tol_prec;   // Function Precision
  double tol_optim;  // optimaliry tolerance
  double min_feas;   // Minor Feasibility Tolerance
  int maxit;         // Major Iteration Limit
  int minit;         // Minor Iteration Limit
  int minittotal;    // Iteration Limit (tummed minor iterations)
  int printlev;      // Print Level
  int mprintlev;     // minor print level
  int superbasics;   // Superbasics Limit
  int hesupdates;    // Hessian Updates
} e04ugf_parameters;

// Ampl Solver Library (ASL) data structure
// (doesn't need to be global but it makes life easier...)
static ASL *asl=0;

// to store any deviations from default opt. settings, -1 to indicate no change
static e04ugf_parameters param = {-1, -1, -1, -1, -1, -1, -1, 5, -1, -1, -1}; 

// definition of keywords AMPL should react to
static keyword keywds[] = {
    KW("feasibility",D_val, &param.tol_feas,	"Major Feasibility Tolerance (default sqrt(eps))"),
    KW("hupdates", I_val, &param.hesupdates, "Hessian Updates (default 20 or Inf)"),
    KW("maxit",	   I_val, &param.maxit, "Major Iteration Limit (default 1000)"),
    KW("minfeas",D_val, &param.min_feas,	"Minor Feasibility Tolerance (default sqrt(eps))"),
    KW("minit",	     I_val, &param.minit, "Minor Iteration Limit (default 500)"),
    KW("minittotal", I_val, &param.minittotal, "Iteration Limit - summed minor iterations (default 10000)"),
    KW("minprint",   I_val, &param.mprintlev, "Minor Print Level (default 0)"),
    KW("optimality", D_val, &param.tol_optim,	"Optimality Tolerance (default sqrt(eps))"),
    KW("precision",	 D_val, &param.tol_prec,	"Function Precision (default eps^0.8)"),
    KW("print",      I_val, &param.printlev, "Print Level (default 5)"),
    KW("superbasics",I_val, &param.superbasics, "Superbasics Limit (default min(500,n+1))"),
  };

// usage message to be printed for '-?'
static char *usage_msg[] = {
  "  where  stub  is a file created from  `ampl -obstub`  or  `ampl -ogstub`",
  "  call.  Assignments have the form  `keyword=value`. To get a list of all",
  "  keywords, call `./e04ug.exe -=`, see description bellow.",
  "  Example:",
  "    ./e04ug.exe problem.nl maxit=100 precision=1e-2 print=0",
  " ",
  "  Alternatively, you can invoke solver E04UGF from Ampl environment. Run",
  "  `ampl` and input the following commands:",
  "     model problem.mod;",
  "     option solver e04ug;",
  "     option e04ug_options 'maxit=100 precision=1e-2 print=1';",
  "     solve;",
  0 };

// possible final messages
static char *end_msg[] = {
  /*0*/ "converged: optimal solution found",
  /*1*/ "failed: the problem is infeasible",
  /*2*/ "failed: the problem seems to be unbounded",
  /*3*/ "failed: the problem might be unbounded,\n  please check optional parameters Unbounded Objective and Unbounded Step Size",
  /*4*/ "failed: maximum number of iterations reached",
  /*5*/ "converged: suboptimal solution found, accuracy couldn't be achieved",
  /*6*/ "error: superbasics limit is too small, increase it using option 'superbasics'",
  /*7*/ "error: please consult the manual"
  };

// info structure of the solver, messages, option settings
static Option_Info Oinfo = {"E04UG", "E04UG, NAG Fortran Library","e04ug_options", keywds, nkeywds, 0 /*funcadd?*/, "E04UG, NAG Fortran Library, AMPL interface 0.9 (beta)", usage_msg, 0, 0, 0 /*cmd line options?*/, 0, 20101111};

// set non-default option settings given by user through AMPL
void set_options(void) {
  char buff[80];
  int len=80;

  // set the level for verification of derivatives, to turn off -> -1
  len=sprintf(buff,"Verify Level = %i ",VERIFYLEVEL);
  E04UJF(buff,len);

  if (param.tol_feas>0) {
    len=sprintf(buff,"Major Feasibility Tolerance = %e ",param.tol_feas);
    E04UJF(buff,len);
  }
  if (param.min_feas>0) {
    len=sprintf(buff,"Minor Feasibility Tolerance = %e ",param.min_feas);
    E04UJF(buff,len);
  }
  if (param.tol_prec>0) {
    len=sprintf(buff,"Function Precision = %e ",param.tol_prec);
    E04UJF(buff,len);
  }
  if (param.tol_optim>0) {
    len=sprintf(buff,"Optimality Tolerance = %e ",param.tol_optim);
    E04UJF(buff,len);
  }
  if (param.maxit!=-1) {
    len=sprintf(buff,"Major Iteration Limit = %i ",param.maxit);
    E04UJF(buff,len);
  }
  if (param.minit!=-1) {
    len=sprintf(buff,"Minor Iteration Limit = %i ",param.minit);
    E04UJF(buff,len);
  }
  if (param.minittotal!=-1) {
    len=sprintf(buff,"Iteration Limit = %i ",param.minittotal);
    E04UJF(buff,len);
  }
  if (param.printlev!=-1) {
    len=sprintf(buff,"Print Level = %i ",param.printlev);
    E04UJF(buff,len);
  }
  if (param.mprintlev!=-1) {
    len=sprintf(buff,"Minor Print Level = %i ",param.mprintlev);
    E04UJF(buff,len);
  }
  if (param.superbasics!=-1) {
    len=sprintf(buff,"Superbasics Limit = %i ",param.superbasics);
    E04UJF(buff,len);
  }
  if (param.hesupdates!=-1) {
    len=sprintf(buff,"Hessian Updates = %i ",param.hesupdates);
    E04UJF(buff,len);
  }
}

// function to evaluate constraints by calling AMPL functions
// expects in ruser allocated 2*N double, iuser[1] and further - lookuptable
void NAG_CALL my_constr(int *mode, const int *ncnln, const int *njnln, const int *nnzjac, const double *x, double *f, double *fjac, const int *nstate, int *iuser, double *ruser) {

  double *xlong=ruser;
  double *vtmp=ruser+n_var;
  int *lookuptable=iuser+1;
  int i,j,k;
  cgrad *cg;
  fint nerror=0;

  if (!mode || !ncnln || !njnln || !nnzjac || !x || !f || !fjac || !nstate || !iuser || !ruser)
    return;

  //copy x->xlong with zeros where the variable is linear
  for (i=0; i<*njnln; i++)
    xlong[i]=x[i];
  for (;i<n_var;i++)
    xlong[i]=0.0;

  if (*mode==0 || *mode==2) {  // function value
    for (i=0;i<*ncnln;i++) {
      f[i]=conival(i,xlong,&nerror);
      if (nerror) {
        *mode=-1;  // cannot be evaluated
        return;
      }
    }
  }
  if (*mode==1 || *mode==2) {  // derivative
    k=0; // counter of elements
    for (i=0; i<*ncnln; i++) {
      congrd(i,xlong,vtmp,&nerror);
      if (nerror) {
        *mode=-1;  // cannot be evaluated
        return;
      }
      for (j=0,cg=Cgrad[i];cg;cg=cg->next) {
        if (cg->varno<*njnln)
          fjac[lookuptable[k++]]=vtmp[j++]; 
          //fjac[k++]=vtmp[j++];   // ??? really like that? not ... store by columns
        else
          break;
      }
    }
  }
}

// function to evaluate value + derivative of the objective function by AMPL
// expects allocated 2*N double in ruser; iuser[0]=nobj
void NAG_CALL my_obj(int *mode, const int *nonln, const double *x, double *objf, double *objgrd, const int *nstate, int *iuser, double *ruser) {

  int i;
  double *xlong=ruser, *vtmp=ruser+n_var;
  int nobj;
  fint nerror=0;

  if (!mode || !nonln || !x || !objf || !objgrd || !nstate || !iuser || !ruser)
    return;

  nobj=iuser[0];

  // copy x to the full vector, zeros anywhere outside nonlin vars
  for (i=0; i<*nonln; i++)
    xlong[i]=x[i];
  for (;i<n_var;i++)
    xlong[i]=0.0;

  // evaluate objective function value
  if (*mode==0 || *mode==2) {
    if (nobj>=0) {
      *objf=objval(nobj,xlong,&nerror);
      if (nerror) {
        *mode=-1;  // cannot be evaluated
        return;
      }
    }
    else  // const. objective ... shouldn't be called but just for sure
      *objf=0;
  }

  // evaluate gradient of the objective
  if (*mode==1 || *mode==2) {
    if (nobj>=0) {
      objgrd(nobj, xlong, vtmp, &nerror);  //vtmp is not compressed
      if (nerror) {
        *mode=-1;  // cannot be evaluated
        return;
      }
      for (i=0;i<*nonln;i++)
        objgrd[i]=vtmp[i];
    }
    else {       // const. objective
      for(i=0;i<*nonln;i++)
        objgrd[i]=0.0;
    }
  }
}

int main(int argc, char *argv[]) {

  FILE *nlfile=0;
  char *stubname;
  char *progname=argv[0];
  int nobj=0;   // number of objective to use, -1 for constant
  int errflag;
  int retflag=0;
  int i,where,neediobj;
  int nnzjnln=0;        //# nonzeroes in nonlinear part of jacobian
  int *lookuptable=0;   // where in A store nonlin. part of Jacobian
  int *kanln=0;         // 'ka' for nonlinear part of Jacobian
  double coef;
  cgrad *cg;
  ograd *og;

  // e04ugf vars
  int n, m, ncnln, nonln, njnln, iobj, nnz, *ha=0, *ka=0, nname, ns, *istate=0, miniz, minz, ninf, *iz=0, leniz, lenz, *iuser=0, ifail;
  double *a=0, *bl=0, *bu=0, *xs=0, *clambda=0, sinf, obj, *z=0, *ruser=0;
  int oldleniz, oldlenz;
  double bigbnd=1e20;


  // prepare our AMPL problem structure to compure function values and 1st deriv
  asl = ASL_alloc(ASL_read_fg);
  // find problem name (stub) & react on -=, -?
  stubname = getstub(&argv, &Oinfo);
  if (!stubname) {
    printf("%s\n",Oinfo.bsname);
    printf("Licence: %s\n\nBased on library:\n", A00ACF() ? "valid" : "valid licence NOT found");
    A00AAF();   // print info about NAG library
    printf("\n");
    printf("No test problem set, terminating.\n");
    printf("Call `%s -?` for help.\n",progname);
    retflag=0;
    goto cleaning;
  }

  return_nofile = 1;
  // load basic characteristics of the problem
  nlfile = jac0dim(stubname, (ftnlen)strlen(stubname));
  if (!nlfile) {
    printf("Can't open %s\nTerminating.\n",stubname);
    retflag=1;
    goto cleaning;
  }

  // is it a "normal" nonlinear continuous problem?
  if (n_var<=0 || nbv>0 || niv>0 || nlnc>0 || lnc>0) {
    printf("Discrete problems or problems with newtork constraints are not supported by this solver.\nTerminating.\n");
    retflag=1;
    goto cleaning;
  }
  if (n_obj<=0) {
    printf("Warning: There is no objective function, pretending to be constant.\n");
    nobj=-1;
  } else if (n_obj>1) {
    printf("Warning: More than one objetive function detected, using the first one.\n");
  }
  
  // AMPL uses their "own" Malloc and type real - only to allocate their bits
  // by allocating these we notify AMPL to fill the data in
  // there are many macros in asl.h which changes many elements of the
  // structure 'asl' such that they look as local variables, for example:
  //    #define n_var   asl->i.n_var_
  //    #define LUv     asl->i.LUv_
  // So if there is no local definition of the variable it is likely to be
  // a part of 'asl' structure
  LUv = (real *)Malloc(n_var*sizeof(real));
  Uvx = (real *)Malloc(n_var*sizeof(real));
  LUrhs = (real *)Malloc(n_con*sizeof(real));
  Urhsx = (real *)Malloc(n_con*sizeof(real));
  X0 = (real *)Malloc(n_var*sizeof(real));
  pi0 = (real *)Malloc(n_con*sizeof(real));
  if (!X0 || !LUv || ! Uvx || (n_con && (!LUrhs || !Urhsx || !pi0))) {
    printf("Memory allocation problem, the problem is too big to fit into the memory. Terminating.\n");
    retflag=1;
    goto cleaning;
  }
  
  // read the rest of the problem in
  if ((errflag=fg_read(nlfile, ASL_return_read_err))!=0) {
    printf("AMPL problem with reading the problem, err_code: %i\nTerminating\n",errflag);
    retflag=1;
    goto cleaning;
  }

  // read additional options and store them into global 'param'
  getopts(argv, &Oinfo);  
  set_options();

  asl->i.congrd_mode = 1;  
     // condensed mode of gradients of constrants (don't store constant zeros)

  // correct min/max: AMPL: objtype[]==0 -> min, ==1 ->max
  if (nobj==-1) {
    // ignore objective function, find feasible point
    // (there is no objective anyway)
    E04UJF("Feasible Point",14);  
  } else if (objtype[nobj]) {
    // minimization is by default --> change it
    E04UJF("Maximize",8);   
  }

  // E04UGF init
  n=n_var;
  m=n_con;   // might need +1 for 'free' IOBJ row
  iobj=0;
  if (nobj!=-1 && nlvo<n_var) {  
    // ok, there are some variables which aren't nonlinear
    // but are they involved in the objective function at all?
    for (neediobj=0,og=Ograd[nobj]; og; og=og->next)
      if(og->varno>=nlvo) {
        neediobj=1;
        break;
      }
    if (neediobj)
      iobj=++m;  // is 1-based, last row
  } 
  if (m==0) {
    m=1;
    iobj=-1;   // 'dummy' free iobj row
  }
  ncnln=nlc;   //# nonlin. constr.
  nonln=nlvo;  //# nonlinear objective variables
  njnln=nlvc;  //# nonlinear constraints variables
  // find the structure of A (Jacobian of constraints with optional IOBJ)
  ka = (int *) calloc(n+1, sizeof(int));
  kanln = (int *) calloc(nlvc+1, sizeof(int));
  if (!ka || !kanln) {
    printf("Memory allocation problem, terminating.\n");
    retflag=1;
    goto cleaning;
  }
  if (iobj==-1) {
    ka[0]=1;
    for (i=1;i<=n;i++)
      ka[i]=2;
    nnz=1;
  }
  else {
    for (i=0;i<n_con;i++)
      for (cg=Cgrad[i];cg;cg=cg->next) {
        ka[cg->varno+1]++;
        if (i<nlc && cg->varno<nlvc)
          kanln[cg->varno+1]++; // count number of nonzeroes in nonlin jac. part
          //nnzjnln++;  // count number of nonzeroes in nonlin jac. part
      }
    if (iobj>0)
      for (og=Ograd[nobj];og;og=og->next)
        ka[og->varno+1]++;
    for (i=1;i<n;i++)   // ka[0]=0 anyway -> skip
      ka[i+1]+=ka[i];   // careful - stil 0-based
    for (i=1;i<nlvc;i++)
      kanln[i+1]+=kanln[i];
    nnz=ka[n];
    nnzjnln=kanln[nlvc];
  }
  nname=1;
  ns=0;
  miniz=0;
  minz=0;
  ninf=0;
  sinf=0.0;
  obj=0.0;
  leniz=500 > n+m ? 500 : n+m;
  lenz=500;
  ifail=-1;
  a = (double *) calloc(nnz, sizeof(double));
  ha = (int *) calloc(nnz, sizeof(int));
  bl = (double *) malloc((n+m)*sizeof(double));
  bu = (double *) malloc((n+m)*sizeof(double));
  xs = (double *) calloc(n+m,sizeof(double));
  istate = (int *) calloc(n+m,sizeof(int));
  clambda = (double *) calloc(n+m, sizeof(double));
  iz = (int *) calloc(leniz,sizeof(int));
  z = (double *) calloc(lenz,sizeof(double));
  iuser = (int *) malloc((nnzjnln+1)*sizeof(int));
  ruser = (double *) malloc(2*n*sizeof(double));

  if (!a || !ha || !bl || !bu || !xs || !istate || !clambda || !iz || !z || !iuser || !ruser) {
    printf("Memory allocation problem, terminating.\n");
    retflag=1;
    goto cleaning;
  }
  iuser[0]=nobj;
  lookuptable=iuser+1;

  if (iobj==-1) {
    // nnz=1 & ka[] is ready
    ha[0]=1;
  }
  else {
    nnzjnln=0;
    for (i=0;i<nlc;i++)  // nonlinear constraints come first
      for (cg=Cgrad[i];cg;cg=cg->next) {
        where=ka[cg->varno]++;
        ha[where]=i+1;       //1-based for fortran
        a[where]=cg->coef;
        if (cg->varno<nlvc)
          lookuptable[nnzjnln++]=kanln[cg->varno]++;
      }
    for (;i<n_con;i++)  // cover the rest - linear
      for (cg=Cgrad[i];cg;cg=cg->next) {
        where=ka[cg->varno]++;
        ha[where]=i+1;       //1-based for fortran
        a[where]=cg->coef;
      }
    if (iobj>0)   // linear part comes last
      for (og=Ograd[nobj];og;og=og->next) {
        if (og->varno<nlvo)
          coef=0;
        else
          coef=og->coef;
        where=ka[og->varno]++;
        ha[where]=iobj;       //last row
        a[where]=coef;
      }

    //repair ka[] & make it 1-based
    for (i=n;i>0;i--)
      ka[i]=ka[i-1]+1;   // 1-based
    ka[0]=1;
  }

  // copy bounds
  for (i=0; i<n;i++) {
    bl[i]=LUv[i];
    bu[i]=Uvx[i];
  }
  for (i=0; i<n_con; i++) {
    bl[i+n]=LUrhs[i];
    bu[i+n]=Urhsx[i];
  }
  if (iobj>0) {   // should be iobj==m, i.e., 1-based
    bl[iobj-1+n]=-bigbnd;
    bu[iobj-1+n]=bigbnd;
  }
  else if (iobj==-1) {  // then m=1
    bl[n]=-bigbnd;
    bu[n]=bigbnd;
  }
  // initial estimate
  for (i=0;i<n;i++)
    xs[i]=X0[i];
  for (i=0;i<ncnln;i++)
    clambda[n+i]=pi0[i];

  while (ifail==-1) {

    E04UGF(my_constr,my_obj,&n,&m,&ncnln,&nonln,&njnln,&iobj,&nnz,
         a,ha,ka,bl,bu,"C" /*start*/,1,&nname,"dummy" /*names*/,1,&ns,xs,
         istate,clambda,&miniz,&minz,&ninf,&sinf,&obj,iz,
         &leniz,z,&lenz,iuser,ruser,&ifail);

    if (ifail==12 || ifail==13 || ifail==15 || ifail==16) {  // not enough mem
      oldleniz=leniz;
      oldlenz=lenz;

      leniz = leniz < miniz ? 1.5*miniz : leniz;
      leniz = (ifail==12 || ifail==15) && leniz<2*oldleniz ? 2*oldleniz : leniz; 
      lenz = lenz < minz ? 1.5*minz : lenz;
      lenz = (ifail==13 || ifail==16) && lenz<2*oldlenz ? 2*oldlenz : lenz;

      if (leniz>oldleniz)
        iz = (int *) realloc(iz,leniz*sizeof(int));
      if (lenz>oldlenz)
        z = (double *) realloc(z,lenz*sizeof(double));

      if (!iz || !z) {
        printf("Memory allocation problem, terminating.\n");
        retflag=1;
        goto cleaning;
      }
      else
        printf("\nMemory allocated, re-running the problem.\n");

      ifail=-1;  // restart & try again
    }
  }

  // results check & save/store results
  if (ifail)
    printf("E04UG finished with an error/warning; flag: %i\n",ifail);
  else
    printf("Result OK\n");

  // write results to AMPL
  for (i=0;i<n_var;i++)
    X0[i]=xs[i];
  for (i=0;i<n_con;i++)
    pi0[i]=clambda[n+i];
  // report the solution back to AMPL
  write_sol( (ifail>=0 && ifail<7) ? end_msg[ifail] : end_msg[7],X0,pi0,&Oinfo);

cleaning: ;  // something to clean?
  if (ka) free(ka);
  if (kanln) free(kanln);
  if (a) free(a);
  if (ha) free(ha);
  if (bl) free(bl);
  if (bu) free(bu);
  if (xs) free(xs);
  if (istate) free(istate);
  if (clambda) free(clambda);
  if (iz) free(iz);
  if (z) free(z);
  if (iuser) free(iuser);
  if (ruser) free(ruser);

  if (asl)
    ASL_free(&asl);
      
  return retflag;
}

